/**
 * Copyright 2012 Bull S.A.S.
 *
 * Licensed  under the  Apache License,  Version 2.0  (the "License");
 * you may not use  this file  except in  compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.jasmine.mule.providers.jms;

import org.mule.providers.jms.JmsConnector;
import org.mule.transaction.TransactionCoordination;
import org.mule.umo.TransactionException;
import org.mule.umo.UMOTransaction;

import javax.jms.JMSException;
import javax.jms.Session;
import java.util.ArrayList;


/**
 * Implementation for Jasmine:
 * Implements a cache for sessions used outside transaction.
 * We use this class because it has been described in jms.properties
 */

public class JasmineJmsConnector extends JmsConnector {

    private ArrayList<Session> sessionCache = new ArrayList<Session>();
    
    private int count = 0;

    public JasmineJmsConnector() {
        super();
        logger.debug("");
    }

    public Session getSession(boolean transacted, boolean topic) throws JMSException {
        if (!isConnected()) {
            throw new JMSException("Not connected");
        }

        Session session = getSessionFromTransaction();
        if (session != null) {
            logger.debug("session retrieved from tx");
            return session;
        }

        // Get Session from the cache, or create a new one.
        synchronized (sessionCache) {
            // close all sessions in the cache every 5000 calls
            // Not sure this is still needed: Was to fight a memory leak.
            if (++count % 5000 == 0) {
                logger.warn("Close all sessions from cache: " + sessionCache.size());
                while (!sessionCache.isEmpty()) {
                    session = sessionCache.remove(0);
                    session.close();
                }
                session = null;
            }
            if (!sessionCache.isEmpty()) {
                session = sessionCache.remove(0);
                if (logger.isDebugEnabled()) {
                    logger.debug("session " + session + " retrieved from cache " + sessionCache.size());
                }
            }
        }
        if (session == null) {
            session = getJmsSupport().createSession(getConnection(), topic, transacted, getAcknowledgementMode(), isNoLocal());
            if (logger.isDebugEnabled()) {
                logger.debug("new session created " + session);
            }
        }

        // bind session to current transaction
        UMOTransaction tx = TransactionCoordination.getInstance().getTransaction();
        if (tx != null) {
            if (logger.isDebugEnabled()) {
                logger.debug("Binding session " + session + " to current transaction " + tx);
            }
            try {
                tx.bindResource(getConnection(), session);
            } catch (TransactionException e) {
                closeQuietly(session);
                throw new RuntimeException("Could not bind session to current transaction", e);
            }
        }
        return session;
    }

    /**
     * When a Session is closed, it is put in the cache for later use.
     *
     * @param session
     * @throws JMSException
     */
    public void close(Session session) throws JMSException {
        if (session != null) {
            synchronized (sessionCache) {
                sessionCache.add(session);
            }
            if (logger.isDebugEnabled()) {
                logger.debug("Return Session " + session + " in the cache " + sessionCache.size());
            }
        }
    }


}
