/**
 * Copyright 2012 Bull S.A.S.
 *
 * Licensed  under the  Apache License,  Version 2.0  (the "License");
 * you may not use  this file  except in  compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.jasmine.mule.providers.jms.transformers;

import org.mule.impl.RequestContext;
import org.mule.providers.jms.JmsConnector;
import org.mule.providers.jms.JmsMessageUtils;
import org.mule.providers.jms.transformers.AbstractJmsTransformer;
import org.mule.transaction.TransactionCoordination;
import org.mule.umo.UMOEventContext;
import org.mule.umo.UMOTransaction;
import org.mule.umo.transformer.TransformerException;

import javax.jms.Message;
import javax.jms.Session;

public abstract class JasmineAbstractJmsTransformer extends AbstractJmsTransformer {

    /**
     * Transform an Object to a JMS Message
     * Actually, there is no need to change this for Jasmine because we do use a pool of sessions.
     * @param src Object to transform
     * @return  the JMS Message
     * @throws TransformerException
     */
    protected Message transformToMessage(Object src) throws TransformerException {
        Session session = null;
        try {
            Message result;

            if (src instanceof Message) {
                result = (Message) src;
                result.clearProperties();
            } else {
                session = this.getSession();
                result = JmsMessageUtils.toMessage(src, session);
            }

            // set the event properties on the Message
            UMOEventContext ctx = RequestContext.getEventContext();
            if (ctx == null) {
                logger.warn("There is no current event context");
                return result;
            }

            this.setJmsProperties(ctx.getMessage(), result);

            return result;
        } catch (TransformerException tex) {
            throw tex;
        } catch (Exception e) {
            throw new TransformerException(this, e);
        } finally {
            /*
                Though JmsTransaction has this session.getTransacted() validation already, we're taking extra precautions
                to cover XA cases and potentially to make up for a configuration error. E.g. omitting transaction
                configuration from an outbound endpoint or router. Note, XA support in Mule will deliberately
                fail with fanfares to signal this case, which is really a user error.
              */

            if (session != null) {
                UMOTransaction muleTx = TransactionCoordination.getInstance().getTransaction();

                final JmsConnector connector = (JmsConnector) endpoint.getConnector();
                if (muleTx == null) {
                    if (logger.isDebugEnabled()) {
                        logger.debug("Closing non-transacted jms session: " + session);
                    }
                    connector.closeQuietly(session);
                } else if (!muleTx.hasResource(connector.getConnection())) {
                    // this is some other session from another connection, don't let it leak
                    if (logger.isDebugEnabled()) {
                        logger.debug("Closing an orphaned, but transacted jms session: " + session +
                                ", transaction: " + muleTx);
                    }
                    connector.closeQuietly(session);
                }
            }
            // aggressively killing any session refs
            session = null;
        }
    }

}
